#include <v_lib.h>
#include "math.h"

#define		SPACE_TIME	0
#define		SPACE_FREQ	1

typedef struct cplx_vector_s
{
	u32_t	size;
//	u32_t	current_space;
	vec_p	elements;
	vec_p	im_elements;
}	cplx_vector_t, *cplx_vector_p;

typedef struct cplx_matrix_s
{
	u32_t	cols, rows;
//	u32_t	current_space;
	vec_p	*cell;
	vec_p	*im_cell;
	vec_p	data;
}	cplx_matrix_t, *cplx_matrix_p;

/*
matAllocCplx
matDuplicateCplx
*/

void
matZeroRows(matrix_p m,u32_t row_start,u32_t row_end)
{
	u32_t r, row_len;
	V_ASSERT(m);
	if(row_start > row_end || row_start > m->rows)
		return;
	if(row_end > m->rows)
		row_end = m->rows;
	row_len = m->cols * sizeof(vec_t);
	for(r = row_start;r < row_end;r++)
		memZero(m->cell[r],row_len);
}

void
matOffset(matrix_p m,vec_t offset)
{
	u32_t r, c;
	V_ASSERT(m);
	for(r = 0;r < m->rows;r++)
		for(c = 0;c < m->cols;c++)
			m->cell[r][c] += offset;
}

void
matZeroCols(matrix_p m,u32_t col_start,u32_t col_end)
{
	V_ASSERT(m);
	if(m->rows == m->cols)
	{
		matTransposeIP(m);
		matZeroRows(m,col_start,col_end);
		matTransposeIP(m);
	}
	else
	{
		matrix_p transposed = matAlloc(m->cols,m->rows);
		matTranspose(transposed,m);
		matZeroRows(transposed,col_start,col_end);
		matTranspose(m,transposed);
		memFree(transposed);
	}
}

void
matBandReject(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	matZeroCols(m,lo_bound,hi_bound);
	matZeroRows(m,lo_bound,hi_bound);
}

void
matBandPass(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	if(lo_bound)
		matBandReject(m,0,lo_bound-1);
	if(hi_bound < m->rows-1)
		matZeroRows(m,hi_bound+1,m->rows-1);
	if(hi_bound < m->cols-1)
		matZeroCols(m,hi_bound+1,m->cols-1);
}

void
vecRunningSum(vector_p dest,vector_p src)
{
	u32_t i;
    vec_p d = dest->elements, s = src->elements;
	V_ASSERT(src->size == dest->size);
    d[0] = s[0];
    for(i = 1;i < dest->size;i++)
    	d[i] = s[i] + d[i-1];
}

void
vecLinSpaced(vector_p dest,vec_t min,vec_t max)
{
	u32_t i, n_divs;
    vec_t delta;
    n_divs = dest->size;
    delta = (max - min) / (n_divs - 1);
    dest->elements[0] = min;
    for(i = 1;i < n_divs;i++)
    	dest->elements[i] = dest->elements[i-1] + delta;
}

void
matLinSpacedRows(matrix_p dest,vec_t min,vec_t max)
{
	u32_t i, n_divs;
    vec_t delta;
    n_divs = dest->cols;
    V_ASSERT(n_divs > 1);
    delta = (max - min) / (n_divs - 1);
    dest->cell[0][0] = min;
    for(i = 1;i < dest->cols;i++)
    	dest->cell[0][i] = dest->cell[0][i-1] + delta;
    for(i = 1;i < dest->rows;i++)
    	memCopy(&(dest->cell[i][0]),&(dest->cell[0][0]),n_divs * sizeof(vec_t));
}

void
matLinSpacedCols(matrix_p dest,vec_t min,vec_t max)
{
	u32_t i, j, n_divs;
    vec_t delta;
    n_divs = dest->rows;
    V_ASSERT(n_divs > 1);
    delta = (max - min) / (n_divs - 1);
    dest->cell[0][0] = min;
    for(i = 1;i < dest->rows;i++)
    	dest->cell[i][0] = dest->cell[i-1][0] + delta;
    for(i = 1;i < dest->cols;i++)
    {
    	for(j = 0;j < dest->rows;j++)
    		dest->cell[j][i] = dest->cell[j][0];
    }
}

void
matScaleAdd(matrix_p dest,matrix_p src,vec_t scale,vec_t offset)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = offset + src->cell[r][c] * scale;
}

void
vecScaleAdd(vector_p dest,vector_p src,vec_t scale,vec_t offset)
{
	u32_t c;
	V_ASSERT(dest->size == src->size);
	for(c = 0;c < dest->size;c++)
		dest->elements[c] = offset + src->elements[c] * scale;
}

void
matScaleM(matrix_p dest,matrix_p src,matrix_p scale)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	V_ASSERT(dest->rows == scale->rows);
	V_ASSERT(dest->cols == scale->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = src->cell[r][c] * scale->cell[r][c];
}

void
matOffsetM(matrix_p dest,matrix_p src,matrix_p offset)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	V_ASSERT(dest->rows == offset->rows);
	V_ASSERT(dest->cols == offset->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = src->cell[r][c] + offset->cell[r][c];
}

void vecExpandToCplx(vector_p dest,vec_p src)
{
	u32_t i;
	for(i = 0;i < dest->size;)
	{
		dest->elements[i++] = *src++;
		dest->elements[i++] = 0;
	}
}

void vecInterleaveCplx(vector_p dest,vec_p re_src,vec_p im_src)
{
	u32_t i;
	for(i = 0;i < dest->size;)
	{
		dest->elements[i++] = *re_src++;
		dest->elements[i++] = *im_src++;
	}
}

void vecSeparateCplx(vec_p re_dest,vec_p im_dest,vector_p src)
{
	u32_t i;
	for(i = 0;i < src->size;)
	{
		*re_dest++ = src->elements[i++];
		*im_dest++ = src->elements[i++];
	}
}

void vecExtractReal(vec_p re_dest,vector_p src)
{
	u32_t i;
	for(i = 0;i < src->size;i += 2)
		*re_dest++ = src->elements[i];
}

V_LOCAL u32_t
int_log2(u32_t val)
{
	u32_t pwr = 1;
	V_ASSERT(IS_POWER_TWO(val));
	while((u32_t)(1 << pwr) < val)
		pwr++;
	return pwr;
}

void vecSeparateCCJ(u32_t ccj_numreals,vec_p re_dest,vec_p im_dest,vec_p ccj)
{
	u32_t i;
	vec_p re_end = re_dest + ccj_numreals;
	vec_p im_end = im_dest + ccj_numreals;
	*re_dest++ = ccj[0];
	*im_dest++ = -ccj[1];
	for(i = 2;i < ccj_numreals;i += 2)
	{
		*re_dest++ = ccj[i];
		*im_dest++ = -ccj[i+1];
		*--re_end = ccj[i];
		*--im_end = ccj[i+1];
	}
}

void	V_CALL
fftForward2DFromReal(matrix_p re,matrix_p im,matrix_p src)
{
	u32_t i, n = src->rows;
	vector_p cplx = vecAlloc(n * 2);
	vhandle_t ctx = fftCreateContext(int_log2(n),false);
	vhandle_t ctx_re = fftCreateContext(int_log2(n),true);
	/* real transform of each row */
	for(i = 0;i < src->rows;i++)
	{
		memCopy32(cplx->elements,src->cell[i],(n * sizeof(vec_t)) >> 2);
		fftForward(ctx_re,cplx->elements);
		vecSeparateCCJ(n,re->cell[i],im->cell[i],cplx->elements);
	}
	/* complex transform of each column */
	matTransposeIP(re);
	matTransposeIP(im);
	for(i = 0;i < src->rows;i++)
	{
		vecInterleaveCplx(cplx,re->cell[i],im->cell[i]);
		fftForward(ctx,cplx->elements);
		vecSeparateCplx(re->cell[i],im->cell[i],cplx);
	}
	fftDestroyContext(ctx);
	fftDestroyContext(ctx_re);
}

void	V_CALL
fftInverse2DToReal(matrix_p dest,matrix_p re,matrix_p im)
{
	u32_t i;
	vector_p cplx = vecAlloc(re->rows * 2);
	vhandle_t ctx = fftCreateContext(int_log2(re->rows),false);
	V_ASSERT(ctx);
	/* inverse the transform of each column */
	for(i = 0;i < re->rows;i++)
	{
		vecInterleaveCplx(cplx,re->cell[i],im->cell[i]);
		fftInverse(ctx,cplx->elements);
		vecSeparateCplx(re->cell[i],im->cell[i],cplx);
	}
	/* inverse the transform of each row */
	matTransposeIP(re);
	matTransposeIP(im);
	for(i = 0;i < re->rows;i++)
	{
		vecInterleaveCplx(cplx,re->cell[i],im->cell[i]);
		fftInverse(ctx,cplx->elements);
		vecExtractReal(dest->cell[i],cplx);
	}
}

void
matBandReject2(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	u32_t neg_lower, neg_upper;
	/*
	 * Assume 'm' is a result of complex Fourier transform.
	 */
	/* Zero positive frequencies. */
	matZeroCols(m,lo_bound,hi_bound);
	matZeroRows(m,lo_bound,hi_bound);
	/* Zero negative frequencies. */
	neg_lower = m->cols - hi_bound;
	neg_upper = m->cols - lo_bound;
	if(neg_upper >= m->cols)
		neg_upper = m->cols - 1;
	if(neg_lower <= neg_upper)
		matZeroCols(m,neg_lower,neg_upper);
	neg_lower = m->rows - hi_bound;
	neg_upper = m->rows - lo_bound;
	if(neg_upper >= m->rows)
		neg_upper = m->rows - 1;
	if(neg_lower <= neg_upper)
		matZeroRows(m,neg_lower,neg_upper);
}

void
matBandPass2(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	u32_t r_limit, c_limit;
	if(m->cols <= 1 || m->rows <= 1)
		return;
	r_limit = (m->rows >> 1) - 1;
	c_limit = (m->cols >> 1) - 1;
	if(lo_bound)
		matBandReject2(m,0,lo_bound - 1);
	if(hi_bound < r_limit)
		matZeroRows(m,hi_bound + 1,m->rows - hi_bound);
	if(hi_bound < c_limit)
		matZeroCols(m,hi_bound + 1,m->cols - hi_bound);
}
