#include <v_lib.h>

/*
 * For given rectangular grid of Z values, perform N passes of
 * bilinear-interpolated upsamplings.
 *
 * Usage:
 *	upsmp <input_file_name> <int num_of_upsampling_passes>
 *
 * The format of input file:
 *	<float X> <float Y> <float Z>
 *	...
 *
 * The 'X' and 'Y' fields must form valid grid, for example:
 *	1 2	100
 *	1 5 200
 *	1 9 300
 *	2 2 500
 *	2 5 600
 *	2 9 700
 * and
 *	1 2	100
 *	2 2 500
 *	1 5 200
 *	2 5 600
 *	1 9 300
 *	2 9 700
 * are both valid (and defines the same grid of Z values).
 */

vboolean row_major;

vector_p X;
vector_p Y;
matrix_p Z;

#define	IsEqual(a,b) (a == b)

vboolean load_source(char *file_name)
{
	u32_t i, j, count, num_pts, num_rows, num_cols;
	vec_p ptr;
	vec3_t pt;
	bstack_t pts;
	vhandle_t scanner;
	stackInit(&pts,sizeof(vec3_t));
	scanner = tokParseFile(file_name,NULL);
	for(;;)
	{
		if(tokNextFloat(scanner,&pt[0]) &&
		   tokNextFloat(scanner,&pt[1]) &&
		   tokNextFloat(scanner,&pt[2]))
			stackPush(&pts,pt);
	}
	tokDestroy(scanner);
	num_pts = stackCount(&pts);
	if(num_pts < 4) /* not enought data */
		return false;
	ptr = (vec_p)stackGetStorage(&pts);
	/* determine order of increments, row/column majority and dimensions */
	num_rows = num_cols = 0;
	for(i = 1;i < num_pts && IsEqual(ptr[0],ptr[3*i]);i++);
	if(i > 1)
	{
		row_major = true;
		num_rows = i - 1;
		num_cols = num_pts / num_rows;
	}
	else
	{
		row_major = false;
		for(i = 1;i < num_pts && IsEqual(ptr[1],ptr[1+3*i]);i++);
		if(i > 1)
		{
			num_cols = i - 1;
			num_rows = num_pts / num_cols;
		}
	}
	if(num_rows * num_cols != num_pts)
	{
		puts("wrong layout of input file");
		return false;
	}
	/* read x, y vectors */
	X = vecAlloc(num_cols);
	Y = vecAlloc(num_rows);
	debugme
	if(row_major)
	{
		for(i=0;i < num_rows;i++)
			Y->elements[i] = ptr[3 * i + 1];
		for(i=0;i < num_cols;i++)
			X->elements[i] = ptr[3 * i * num_rows];
	}
	else
	{
		for(i=0;i < num_cols;i++)
			X->elements[i] = ptr[3 * i];
		for(i=0;i < num_rows;i++)
			Y->elements[i] = ptr[3 * i * num_cols + 1];
	}
	/* read z matrix */
	Z = matAlloc(num_rows,num_cols);
	count = 0;
	/* fixme - implement x/y consistency check */
	ptr += 2;
	if(row_major)
	{
		for(i=0;i < num_cols;i++)
			for(j=0;j < num_rows;j++)
			{
				Z->cell[j][i] = *ptr;
				ptr += 3;
			}
	}
	else
	{
		for(j=0;j < num_rows;j++)
			for(i=0;i < num_cols;i++)
			{
				Z->cell[j][i] = *ptr;
				ptr += 3;
			}
	}
	stackClear(&pts);
	return true;
}

void save_data(char *outfile_name)
{
	u32_t i, j;
	vhandle_t out = fileOpenWrite(outfile_name);
	if(row_major)
	{
		for(i=0;i < Z->cols;i++)
			for(j=0;j < Z->rows;j++)
				filePrintf(out,"%g\t%g\t%g\n",
					X->elements[i], Y->elements[j], Z->cell[j][i]);
	}
	else
	{
		for(j=0;j < Z->rows;j++)
			for(i=0;i < Z->cols;i++)
				filePrintf(out,"%g\t%g\t%g\n",
					X->elements[i], Y->elements[j], Z->cell[j][i]);
	}
	fileClose(out);
}

void upsample(void)
{
	vector_p up_x, up_y;
	matrix_p up_z;
	
	/* ... todo */
}

int main()
{
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	
	if(load_source("gr_src.txt"))
		save_data("out");
	
	if(X)
		memFree(X);
	if(Y)
		memFree(Y);
	if(Z)
		memFree(Z);

	memPrintStats();
	return 0;
}
