#include <v_lib.h>

#define PWR			9
#define	DIM			(1 << PWR)
#define	MAX_ITER	10

matrix_p load_image(char *name)
{
	u32_t w,h,i,j;
	matrix_p m;
	vec_t scale = 1.0 / 3 / 256;
	u8_t *ptr,*img = tgaLoad(name,&w,&h);
	if(!img)
		return NULL;
	m = matAlloc(w,h);
	ptr = img;
	for(i=0;i<m->rows;i++)
		for(j=0;j<m->cols;j++)
		{
			m->cell[i][j] = (ptr[0]+ptr[1]+ptr[2]) * scale;
			ptr += 4;
		}
	memFree(img);
	return m;
}

void save_image(char *name,matrix_p src)
{
	u32_t i,j;
	vec_t val;
	u8_t *ptr, *img = (u8_p)memAllocZ(src->rows * src->cols * 4);
	if(!img)
		return;
	ptr = img;
	for(i=0;i < src->rows;i++)
		for(j=0;j < src->cols;j++)
		{
			val = src->cell[i][j] * 255;
			if(val < 0)
				val = 0;
			else if(val > 255)
				val = 255;
			ptr[0] = ptr[1] = ptr[2] = (u8_t)val;
			ptr += 4;
		}
	bmpSave(name,src->cols,src->rows,img);
	memFree(img);
}

vec_t mat_diff(matrix_p m1,matrix_p m2)
{
	u32_t i,j;
	vec_t t,d = 0;
	V_ASSERT(m1->rows==m2->rows && m1->cols==m2->cols);
	for(i=0;i<m1->rows;i++)
		for(j=0;j<m1->cols;j++)
		{
			t = v_abs(m1->cell[i][j] - m2->cell[i][j]);
			if(t > d)
				d = t;
		}
	return d;
}

void
matZeroRows(matrix_p m,u32_t row_start,u32_t row_end)
{
	u32_t r, row_bytes;
	V_ASSERT(m);
	if(row_start > row_end || row_start >= m->rows)
		return;
	if(row_end >= m->rows)
		row_end = m->rows - 1;
	row_bytes = m->cols * sizeof(vec_t);
	for(r = row_start;r <= row_end;r++)
		memZero(m->cell[r],row_bytes);
}

void
matZeroCols(matrix_p m,u32_t col_start,u32_t col_end)
{
	u32_t c, r;
	V_ASSERT(m);
	if(col_start > col_end || col_start >= m->cols)
		return;
	if(col_end >= m->cols)
		col_end = m->cols - 1;
	for(c = col_start;c <= col_end;c++)
		for(r = 0;r < m->rows;r++)
			m->cell[r][c] = 0;
}

void
matScaleAdd(matrix_p dest,matrix_p src,vec_t scale,vec_t offset)
{
	u32_t r, c;
	V_ASSERT(dest->rows == src->rows);
	V_ASSERT(dest->cols == src->cols);
	for(r = 0;r < dest->rows;r++)
		for(c = 0;c < dest->cols;c++)
			dest->cell[r][c] = offset + src->cell[r][c] * scale;
}

void
matBandReject(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	matZeroCols(m,lo_bound,hi_bound);
	matZeroRows(m,lo_bound,hi_bound);
}

void
matBandReject2(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	u32_t neg_lower, neg_upper;
	/*
	 * Assume 'm' is a result of complex Fourier transform.
	 */
	/* Zero positive frequencies. */
	matZeroCols(m,lo_bound,hi_bound);
	matZeroRows(m,lo_bound,hi_bound);
	/* Zero negative frequencies. */
	neg_lower = m->cols - hi_bound;
	neg_upper = m->cols - lo_bound;
	if(neg_upper >= m->cols)
		neg_upper = m->cols - 1;
	if(neg_lower <= neg_upper)
		matZeroCols(m,neg_lower,neg_upper);
	neg_lower = m->rows - hi_bound;
	neg_upper = m->rows - lo_bound;
	if(neg_upper >= m->rows)
		neg_upper = m->rows - 1;
	if(neg_lower <= neg_upper)
		matZeroRows(m,neg_lower,neg_upper);
}

void
matBandPass(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	if(lo_bound)
		matBandReject(m,0,lo_bound-1);
	if(hi_bound < m->rows-1)
		matZeroRows(m,hi_bound+1,m->rows-1);
	if(hi_bound < m->cols-1)
		matZeroCols(m,hi_bound+1,m->cols-1);
}

void
matBandPass2(matrix_p m,u32_t lo_bound,u32_t hi_bound)
{
	u32_t r_limit, c_limit;
	if(m->cols <= 1 || m->rows <= 1)
		return;
	r_limit = (m->rows >> 1) - 1;
	c_limit = (m->cols >> 1) - 1;
	if(lo_bound)
		matBandReject2(m,0,lo_bound - 1);
	if(hi_bound < r_limit)
		matZeroRows(m,hi_bound + 1,m->rows - hi_bound);
	if(hi_bound < c_limit)
		matZeroCols(m,hi_bound + 1,m->cols - hi_bound);
}

vboolean
cspaceAdjustContrast(matrix_p dest,matrix_p src,vec_t min,vec_t max)
{
	vec_t rscale;
    rscale = max - min;
    if(!IsZero(rscale))
    {
    	rscale = 1 / rscale;
    	matScaleAdd(dest,src,rscale,- min * rscale);
    	return true;
    }
   	conlPuts(LOG_WARNING,"cspaceAdjustContrast(): zero input range, can't transform");
    return false;
}

vboolean
cspaceMaximizeContrast(matrix_p m)
{
	u32_t r, c;
	vec_t min = 0, max = 0;
    matMin(m,&r,&c);
    min = m->cell[r][c];
    matMax(m,&r,&c);
    max = m->cell[r][c];
    return cspaceAdjustContrast(m,m,min,max);
}

void vecExpandToCplx(vector_p dest,vec_p src)
{
	u32_t i;
	for(i = 0;i < dest->size;)
	{
		dest->elements[i++] = *src++;
		dest->elements[i++] = 0;
	}
}

void vecInterleaveCplx(vector_p dest,vec_p re_src,vec_p im_src)
{
	u32_t i;
	for(i = 0;i < dest->size;)
	{
		dest->elements[i++] = *re_src++;
		dest->elements[i++] = *im_src++;
	}
}

void vecSeparateCplx(vec_p re_dest,vec_p im_dest,vector_p src)
{
	u32_t i;
	for(i = 0;i < src->size;)
	{
		*re_dest++ = src->elements[i++];
		*im_dest++ = src->elements[i++];
	}
}

void vecExtractReal(vec_p re_dest,vector_p src)
{
	u32_t i;
	for(i = 0;i < src->size;i += 2)
		*re_dest++ = src->elements[i];
}

V_LOCAL u32_t
int_log2(u32_t val)
{
	u32_t pwr = 1;
	V_ASSERT(IS_POWER_TWO(val));
	while((u32_t)(1 << pwr) < val)
		pwr++;
	return pwr;
}

void vecSeparateCCJ(u32_t ccj_numreals,vec_p re_dest,vec_p im_dest,vec_p ccj)
{
	u32_t i;
	vec_p re_end = re_dest + ccj_numreals;
	vec_p im_end = im_dest + ccj_numreals;
	*re_dest++ = ccj[0];
	*im_dest++ = -ccj[1];
	for(i = 2;i < ccj_numreals;i += 2)
	{
		*re_dest++ = ccj[i];
		*im_dest++ = -ccj[i+1];
		*--re_end = ccj[i];
		*--im_end = ccj[i+1];
	}
}

void	V_CALL
fftForward2DFromReal(matrix_p re,matrix_p im,matrix_p src)
{
	u32_t i, n = src->rows;
	vector_p cplx = vecAlloc(n * 2);
	vhandle_t ctx = fftCreateContext(int_log2(n),false);
	vhandle_t ctx_re = fftCreateContext(int_log2(n),true);
	/* real transform of each row */
	for(i = 0;i < src->rows;i++)
	{
		memCopy32(cplx->elements,src->cell[i],(n * sizeof(vec_t)) >> 2);
		fftForward(ctx_re,cplx->elements);
		vecSeparateCCJ(n,re->cell[i],im->cell[i],cplx->elements);
	}
	/* complex transform of each column */
	matTransposeIP(re);
	matTransposeIP(im);
	for(i = 0;i < src->rows;i++)
	{
		vecInterleaveCplx(cplx,re->cell[i],im->cell[i]);
		fftForward(ctx,cplx->elements);
		vecSeparateCplx(re->cell[i],im->cell[i],cplx);
	}
	fftDestroyContext(ctx);
	fftDestroyContext(ctx_re);
}

void	V_CALL
fftInverse2DToReal(matrix_p dest,matrix_p re,matrix_p im)
{
	u32_t i;
	vector_p cplx = vecAlloc(re->rows * 2);
	vhandle_t ctx = fftCreateContext(int_log2(re->rows),false);
	V_ASSERT(ctx);
	/* inverse the transform of each column */
	for(i = 0;i < re->rows;i++)
	{
		vecInterleaveCplx(cplx,re->cell[i],im->cell[i]);
		fftInverse(ctx,cplx->elements);
		vecSeparateCplx(re->cell[i],im->cell[i],cplx);
	}
	/* inverse the transform of each row */
	matTransposeIP(re);
	matTransposeIP(im);
	for(i = 0;i < re->rows;i++)
	{
		vecInterleaveCplx(cplx,re->cell[i],im->cell[i]);
		fftInverse(ctx,cplx->elements);
		vecExtractReal(dest->cell[i],cplx);
	}
}






V_LOCAL void
matUnpackCplxCol(vector_p cplx,matrix_p data,u32_t col)
{
	u32_t i, n = data->rows;
	vec_p ptr = cplx->elements, *cell = data->cell;
	V_ASSERT(cplx->size == data->rows * 2);
	V_ASSERT(!(col & 1));
	for(i = 0;i < n;i++)
	{
		*ptr++ = cell[i][col];
		*ptr++ = cell[i][col+1];
	}
}

V_LOCAL	void
matPackCplxCol(matrix_p data,vector_p cplx,u32_t col)
{
	u32_t i, n = data->rows;
	vec_p ptr = cplx->elements, *cell = data->cell;
	V_ASSERT(cplx->size == data->rows * 2);
	V_ASSERT(!(col & 1));
	for(i = 0;i < n;i++)
	{
		cell[i][col] = *ptr++;
		cell[i][col+1] = *ptr++;
	}
}

/*
 * Though the output of this function contains all necessary information
 * to restore the original (real-valued) matrix, the processing logic and
 * placement of data have to be explained.
 *
 * At call, 'rvm' contains real-valued elements (so the name). The transform
 * is two-pass. At first, we compute real FT of each row; we got _positive_
 * wavenumbers only. Each WN is a complex value and occupies two consecutive
 * columns. Thus, to compute transform in second dimension, we need _complex_
 * FT of each _pair_ of columns (Re(z) in even, Im(z) in odd). It returns
 * both positive (from zero to highest) and negative (from -highest to
 * -nearest_to_zero) wavenumbers.
 * 
 * In each row: only _positive_ frequencies (from Zero to Nyquist);
 * all values are complex (and for that reason there are only
 * _(data->cols / 2)_ elements in each row).
 
 * In each col: both _positive_ and _negative_ frequencies (from Zero in
 * row #0 to Nyquist in row ... and up to - ... in last row) of real FFT's of each row. All values are
 * complex (since ... there are total data->rows _independant_ elements).
 */
 
void	V_CALL
fftForwardRVM(matrix_p rvm)
{
	u32_t i;
	vector_p cplx = vecAlloc(rvm->rows * 2);
	vhandle_t row_transform = fftCreateContext(int_log2(rvm->cols),true);
	vhandle_t col_transform = fftCreateContext(int_log2(rvm->rows),false);
	/* real transform of each row */
	for(i = 0;i < rvm->rows;i++)
		fftForward(row_transform,rvm->cell[i]);
	/* complex transform of each (complex) column */
	for(i = 0;i < rvm->cols;i += 2)
	{
		matUnpackCplxCol(cplx,rvm,i);
		fftForward(col_transform,cplx->elements);
		matPackCplxCol(rvm,cplx,i);
	}
	fftDestroyContext(row_transform);
	fftDestroyContext(col_transform);
}

void	V_CALL
fftInverseRVMFT(matrix_p rvmft)
{
	u32_t i;
	vector_p cplx = vecAlloc(rvmft->rows * 2);
	vhandle_t row_transform = fftCreateContext(int_log2(rvmft->cols),true);
	vhandle_t col_transform = fftCreateContext(int_log2(rvmft->rows),false);
	/* complex transform of each (complex) column */
	for(i = 0;i < rvmft->cols;i += 2)
	{
		matUnpackCplxCol(cplx,rvmft,i);
		fftInverse(col_transform,cplx->elements);
		matPackCplxCol(rvmft,cplx,i);
	}
	/* real transform of each row */
	for(i = 0;i < rvmft->rows;i++)
		fftInverse(row_transform,rvmft->cell[i]);
	fftDestroyContext(row_transform);
	fftDestroyContext(col_transform);
}

#if 0

int main()
{
	matrix_p m, re, im;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);

	m = load_image("img512.tga");
	assert(m);

	re = matAlloc(m->rows,m->cols);
	im = matAlloc(m->rows,m->cols);

	fftForward2DFromReal(re,im,m);
	matBandPass2(re,0,16);
	matBandPass2(im,0,16);
	fftInverse2DToReal(m,re,im);
	

	cspaceMaximizeContrast(m);
	save_image("out.tga",m);

	memFree(m);
	memFree(re);
	memFree(im);
	return 0;
}

#else

matLowPass(matrix_p wn_matrix,u32_t reject_from)
{
	V_ASSERT(reject_from > 0);
	V_ASSERT(reject_from < (wn_matrix->cols >> 1));
	V_ASSERT(wn_matrix->cols == wn_matrix->rows);
	matZeroCols(wn_matrix,reject_from * 2,wn_matrix->cols-1);
	matZeroRows(wn_matrix,reject_from,wn_matrix->rows - reject_from);
}

#if 0

matrix_p
matPSEofRVMFT(matrix_p rvm_transform)
{
	u32_t r, c;
	matrix_p row_pse = matAlloc(rvm_transform->rows >> 1,rvm_transform->cols >> 1);
	matrix_p col_pse = matAlloc(rvm_transform->rows >> 1,rvm_transform->cols >> 1);
	/* PSE of each row (of real FT: freqs from zero to Nyquist) */
	for(r = 0;r < row_pse->rows;r++)
	{
		/* zero wn */
		
	}
	/* PSE of each col (of cplx FT: freqs from zero to -nearest_to_zero) */
	
}

#endif

int main()
{
	matrix_p m;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);

	m = load_image("img512.tga");
	assert(m);

	{
		u32_t rej;
		char name[256];
		matrix_p m1 = matAlloc(m->rows,m->cols);
		for(rej = 2;rej < (m->rows >> 1);rej++)
		{
			matCopy(m1,m);
			matTransposeIP(m1);
			fftForwardRVM(m1);
			matLowPass(m1,rej);
			fftInverseRVMFT(m1);
			cspaceMaximizeContrast(m1);
			sprintf(name,"out%04u.bmp",rej);
			matTransposeIP(m1);
			save_image(name,m1);
		}
	}


	memFree(m);
	return 0;
}

#endif
